#!/usr/bin/python3
#
#   win_app_package_exe_config.py
#
import sys

import ctypes
import ctypes.wintypes

import struct
from namedstruct import namedstruct

is_64bit = (ctypes.sizeof( ctypes.c_voidp ) * 8) == 64

# DumpRange used for debug
struct_DumpRange = namedstruct( 'DumpRange', '<'
    '8h:row1 '
    '8h:row2 '
    '8h:row3 '
    '8h:row4 '
    )

# format of .ICO file on disk
struct_IconDirHeader = namedstruct( 'IconDirHeader', '<'
    'h:idReserved '
    'h:idType '
    'h:idCount'
    )
struct_IconDirEntry = namedstruct( 'IconDirEntry', '<'
    'b:bWidth '
    'b:bHeight '
    'b:bColorCount '
    'b:bReserved '
    'h:wPlanes '
    'h:wBitCount '
    'i:dwBytesInRes '
    'i:dwImageOffset'
    )

# Resource formats
struct_GrpIconDir = namedstruct( 'GrpIconDir', '<'
    'h:idReserved '
    'h:idType '
    'h:idCount'
    )
struct_GrpIconDirEntry = namedstruct( 'GrpIconDirEntry', '<'
    'b:bWidth '
    'b:bHeight '
    'b:bColorCount '
    'b:bReserved '
    'h:wPlanes '
    'h:wBitCount '
    'i:dwBytesInRes '
    'h:nID'
    )

# VS_VERSIONINFO
struct_VersionInfoHeader = namedstruct( 'VersionInfo', '<'
    'h:wLength '    # sizeof( struct_VersionInfoHeader ) + sizeof( struct_VersionInfoFixedFileInfo )
    'h:wValueLength '
    'h:wType '
    '16h:szKey '    # room for "VS_VERSION_INFO\0" in utf-16
    'h:Padding1 '
    )

# VS_FIXEDFILEINFO
struct_VersionInfoFixedFileInfo = namedstruct( 'VersionInfoFixedFileInfo', '<'
    'I:dwSignature '        # magic value 0xFEEF04BD
    'h:dwStrucVersion1 '     # ???
    'h:dwStrucVersion0 '     # ???
    'h:dwFileVersion1 '
    'h:dwFileVersion0 '
    'h:dwFileVersion3 '
    'h:dwFileVersion2 '
    'h:dwProductVersion1 '
    'h:dwProductVersion0 '
    'h:dwProductVersion3 '
    'h:dwProductVersion2 '
    'i:dwFileFlagsMask '
    'i:dwFileFlags '
    'i:dwFileOS '
    'i:dwFileType '
    'i:dwFileSubtype '
    'i:dwFileDateMS '
    'i:dwFileDateLS '
    )
# dwSignature value
VER_SIGNATURE_MAGIC = 0xFEEF04BD
# dwFileOS value
VOS_NT_WINDOWS32    = 0x00040004
# dwFileType
VFT_APP             = 0x00000001
# dwFileSubtype
VFT2_UNKNOWN        = 0x00000000

# Child of VS_VERSION_INFO Header
struct_ChildOfVersionInfoHeader = namedstruct( 'ChildOfVersionInfoHeader', '<'
    'h:wLength '
    'h:wValueLength '
    'h:wType '
    )

# StringFileInfo
struct_StringFileInfo = namedstruct( 'StringFileInfo', '<'
    'h:wLength '
    'h:wValueLength '
    'h:wType '
    '15h:szKey '      # "StringFileInfo".
    )

# VarFileInfo
struct_VarFileInfo = namedstruct( 'VarFileInfo', '<'
    'h:wLength '
    'h:wValueLength '
    'h:wType '
    '12h:szKey '      # "VarFileInfo".
    'h:padding'
    )

struct_Var = namedstruct( 'Var', '<'
    'h:wLength '
    'h:wValueLength '
    'h:wType '
    '11h:szKey '      # "Translation".
    'h:Padding '
    # list of VarValue
    )

struct_VarValue = namedstruct( 'VarValue', '<'
    'I:value '
    )

# StringTable
struct_StringTable = namedstruct( 'StringTable', '<'
    'h:wLength '
    'h:wValueLength '
    'h:wType '
    '15h:szKey '
    'h:Padding '
    # n of String
    )

# String
struct_StringHeader = namedstruct( 'StringHeader', '<'
    'h:wLength '
    'h:wValueLength '
    'h:wType '
    # 0 terminated wchar string
    # pad to 32 bit
    # 0 terminated wchar string
    )


# ctypes functions
BeginUpdateResource = ctypes.windll.kernel32.BeginUpdateResourceW
BeginUpdateResource.argtypes = (ctypes.wintypes.LPCWSTR
                               ,ctypes.wintypes.BOOL)
BeginUpdateResource.restype = ctypes.wintypes.HANDLE

UpdateResource = ctypes.windll.kernel32.UpdateResourceW
if is_64bit:
    UpdateResource.argtypes = (ctypes.wintypes.HANDLE
                              ,ctypes.wintypes.ULARGE_INTEGER
                              ,ctypes.wintypes.ULARGE_INTEGER
                              ,ctypes.wintypes.WORD
                              ,ctypes.wintypes.LPVOID
                              ,ctypes.wintypes.DWORD)
else:
    UpdateResource.argtypes = (ctypes.wintypes.HANDLE
                              ,ctypes.wintypes.DWORD
                              ,ctypes.wintypes.DWORD
                              ,ctypes.wintypes.WORD
                              ,ctypes.wintypes.LPVOID
                              ,ctypes.wintypes.DWORD)

EndUpdateResource = ctypes.windll.kernel32.EndUpdateResourceW
EndUpdateResource.argtypes = (ctypes.wintypes.HANDLE
                             ,ctypes.wintypes.BOOL)

LoadLibraryEx = ctypes.windll.kernel32.LoadLibraryExW
LoadLibraryEx.argtypes =    (ctypes.wintypes.LPCWSTR
                            ,ctypes.wintypes.HANDLE
                            ,ctypes.wintypes.DWORD)
LoadLibraryEx.restype =     ctypes.wintypes.HANDLE

LOAD_LIBRARY_AS_DATAFILE_EXCLUSIVE = 0x00000040

FreeLibrary = ctypes.windll.kernel32.FreeLibrary
FreeLibrary.argtypes = (ctypes.wintypes.HANDLE,)

FindResource = ctypes.windll.kernel32.FindResourceW
FindResource.argtypes =     (ctypes.wintypes.HANDLE
                            ,ctypes.wintypes.ULARGE_INTEGER
                            ,ctypes.wintypes.ULARGE_INTEGER)
FindResource.restype =      ctypes.wintypes.HANDLE

LoadResource = ctypes.windll.kernel32.LoadResource
LoadResource.argtypes =     (ctypes.wintypes.HANDLE
                            ,ctypes.wintypes.HANDLE)
LoadResource.restype =      ctypes.wintypes.HANDLE

SizeofResource = ctypes.windll.kernel32.SizeofResource
SizeofResource.argtypes =   (ctypes.wintypes.HANDLE
                            ,ctypes.wintypes.HANDLE)
SizeofResource.restype =    ctypes.wintypes.DWORD

LockResource = ctypes.windll.kernel32.LockResource
LockResource.argtypes =     (ctypes.wintypes.HANDLE,)
LockResource.restype =      ctypes.c_void_p

RT_ICON =                   3
RT_GROUP_ICON =             14
RT_VERSION =                16
RT_STRING =                 6        # STRINGTABLE

STRINGTABLE_BUNDLE_SIZE =   16

resource_ids = {
    'IDS_PYTHON_DLL':           64*STRINGTABLE_BUNDLE_SIZE + 0,
    'IDS_MAIN_PY_MODULE':       64*STRINGTABLE_BUNDLE_SIZE + 1,
    'IDS_INSTALL_FOLDER_KEY':   64*STRINGTABLE_BUNDLE_SIZE + 2,
    'IDS_INSTALL_FOLDER_VALUE': 64*STRINGTABLE_BUNDLE_SIZE + 3,

    'IDS_PY_VERBOSE':           65*STRINGTABLE_BUNDLE_SIZE + 0,
    }

# called when used standalone
def main( argv ):
    if argv[1:2] == ['bootstrap']:
        exe_filename = argv[2]

        python_dll = argv[3]
        main_py_module = agv[4]
        install_key  = argv[5]
        install_value  = argv[6]

        return configureAppExeBootStrap( exe_filename, python_dll, main_py_module, install_key, install_value )

    elif argv[1:2] == ['flags']:
        exe_filename = argv[2]

        py_verbose = argv[3]

        return configureAppExePyFlags( exe_filename, py_verbose )

    elif argv[1:2] == ['create']:
        return createResourceIdHeaderFile( argv[2] )

    elif argv[1:2] == ['icon']:
        return updateIconInExe( argv[3], argv[2] )

    elif argv[1:2] == ['show-version-info']:
        return showVersionInfoInExe( argv[2] )

    elif argv[1:2] == ['set-version-info']:
        return setVersionInfoInExe( argv[2], argv[3:] )

    else:
        print( 'Usage: %s bootstrap <exefile> <python_dll> <main_py_module> <install_key> <install_value>' % (argv[0],) )
        print( '       %s flags <exefile> <verbose>' % (argv[0],) )
        print( '       %s create <header-filename>' % (argv[0],) )
        print( '       %s show-version-info <exefile>' % (argv[0],) )
        return 1


# called when part of win_app_packager
def flagsCommand( argv ):
    if len(argv) != 4:
        return usage()

    exe_filename = argv[2]

    py_verbose = argv[3]

    return configureAppExePyFlags( exe_filename, py_verbose )

# called when part of win_app_packager
def usage():
################################################################################
    print(
'''python3 -m win_app_packager flags <exe-file> <verbose>'
  exe-file
    - the win_app_package create EXE file to modify
  verbose
    - either "0" or "1". The value to see the python verbose flag to.
''' )
    return 1

def createResourceIdHeaderFile( h_file ):
    with open( h_file, 'w' ) as f:
        for name in resource_ids:
           f.write( '#define %s %d\n' % (name, resource_ids[ name ]) )
    return 0

def configureAppExeBootStrap( exe_filename, python_dll, main_py_module, install_key, install_value ):
    all_strings = [
        python_dll,
        main_py_module,
        install_key,
        install_value,
        ]
    return updateStringBundleInExe( exe_filename, 'IDS_INSTALL_FOLDER_KEY', all_strings )

def configureAppExePyFlags( exe_filename, py_verbose ):
    all_strings = [
        py_verbose,
        ]
    return updateStringBundleInExe( exe_filename, 'IDS_PY_VERBOSE', all_strings )

def updateStringBundleInExe( exe_filename, stringtable_id_name, all_strings ):
    stringtable_id = resource_ids[ stringtable_id_name ]

    while len(all_strings) != STRINGTABLE_BUNDLE_SIZE:
        all_strings.append( '' )

    all_strtab_data = []
    for s in all_strings:
        count = struct.pack( '<H', len( s ) )
        all_strtab_data.append( count )
        if len(s) > 0:
            data = s.encode( 'utf-16' )[2:]
            all_strtab_data.append( data )

    strtab_data = b''.join( all_strtab_data )

    language = 0

    strtab_id = stringtable_id//STRINGTABLE_BUNDLE_SIZE + 1

    h = BeginUpdateResource( exe_filename, False )
    #print( h, ctypes.FormatError() )

    rc = UpdateResource( h, RT_STRING, strtab_id, language, strtab_data, len(strtab_data) )
    #print( rc, ctypes.FormatError() )

    rc = EndUpdateResource( h, False );
    #print( rc, ctypes.FormatError() )

    return 0

def updateIconInExe( exe_filename, icon_filename ):
    with open( icon_filename, 'rb' ) as f:
        all_entries = []
        all_images = []

        header = struct_IconDirHeader.unpack( f.read( len(struct_IconDirHeader) ) )

        for i in range( header.idCount ):
            all_entries.append(
                 struct_IconDirEntry.unpack(
                    f.read( len(struct_IconDirEntry) ) ) )

        for entry in all_entries:
            f.seek( entry.dwImageOffset, 0 )
            all_images.append( f.read( entry.dwBytesInRes ) )

    h = BeginUpdateResource( exe_filename, False )
    #print( h, ctypes.FormatError() )

    grp_header = struct_GrpIconDir.packer()
    grp_header.idReserved = 0
    grp_header.idType = header.idType
    grp_header.idCount = header.idCount

    all_data = [grp_header.pack()]
    entry_id = 1
    for entry in all_entries:
        grp_entry = struct_GrpIconDirEntry.packer()
        grp_entry.bWidth = entry.bWidth
        grp_entry.bHeight = entry.bHeight
        grp_entry.bColorCount = entry.bColorCount
        grp_entry.bReserved = entry.bReserved
        grp_entry.wPlanes = entry.wPlanes
        grp_entry.wBitCount = entry.wBitCount
        grp_entry.dwBytesInRes = entry.dwBytesInRes
        grp_entry.nID = entry_id
        all_data.append( grp_entry.pack() )
        entry_id += 1

    language = 0
    data = b''.join( all_data )
    rc = UpdateResource( h, RT_GROUP_ICON, 1, language, data, len(data) )
    #print( rc, ctypes.FormatError() )

    entry_id = 1
    for image in all_images:
        rc = UpdateResource( h, RT_ICON, entry_id, language, image, len(image) )
        #print( rc, ctypes.FormatError() )
        entry_id += 1

    rc = EndUpdateResource( h, False );
    #print( rc, ctypes.FormatError() )
    return 0

def dprint( msg ):
    #print( '|', msg )
    pass

def showVersionInfoInExe( exe_filename ):
    vir = VersionInfoResource( exe_filename )

    print( '     dwStrucVersion: %d.%d' % (vir.fixed.dwStrucVersion0, vir.fixed.dwStrucVersion1) )
    print( '        FileVersion: %d.%d.%d.%d' % (vir.fixed.dwFileVersion0, vir.fixed.dwFileVersion1, vir.fixed.dwFileVersion2, vir.fixed.dwFileVersion3) )
    print( '     ProductVersion: %d.%d.%d.%d' % (vir.fixed.dwProductVersion0, vir.fixed.dwProductVersion1, vir.fixed.dwProductVersion2, vir.fixed.dwProductVersion3) )

    print( '           encoding: %s' % (vir.encoding,) )
    for key in sorted( vir.all_properties ):
        print( ' %18s: "%s"' % (key, vir.all_properties[ key ]) )

    print( '  translation value: 0x%8.8x' % (vir.translation.value,) )

def setVersionInfoInExe( exe_filename, argv ):
    vir = VersionInfoResource( exe_filename )

    args = iter(argv)

    for name in args:
        value = next(args)

        if name.lower() == 'version':
            vir.setVersion( value )

        else:
            vir.setProperty( name, value )

    vir.updateWithChanges()

class VersionInfoResource:
    def __init__( self, exe_filename ):
        self.exe_filename = exe_filename
        self.__loadVersionInfo()

        self.pack_header = struct_VersionInfoHeader.packer( self.header )
        self.pack_fixed = struct_VersionInfoFixedFileInfo.packer( self.fixed )
        self.pack_child = struct_ChildOfVersionInfoHeader.packer( self.child )
        self.pack_string_file_info = struct_StringFileInfo.packer( self.string_file_info )

    def setProperty( self, name, value ):
        self.all_properties[ name ] = value

    def setVersion( self, int_version ):
        assert len(int_version) == 4, 'bad version %r' % (int_version,)
        str_version = '%d.%d.%d.%d' % int_version

        self.all_properties[ 'ProductVersion' ] = str_version
        self.pack_fixed.dwProductVersion0 = int_version[0]
        self.pack_fixed.dwProductVersion1 = int_version[1]
        self.pack_fixed.dwProductVersion2 = int_version[2]
        self.pack_fixed.dwProductVersion3 = int_version[3]

        self.all_properties[ 'FileVersion' ] = str_version
        self.pack_fixed.dwFileVersion0 = int_version[0]
        self.pack_fixed.dwFileVersion1 = int_version[1]
        self.pack_fixed.dwFileVersion2 = int_version[2]
        self.pack_fixed.dwFileVersion3 = int_version[3]

    def updateWithChanges( self ):
        version_resource = self.__encodeVersionInfo()
        self.__decodeVersionInfo( version_resource )

        h = BeginUpdateResource( self.exe_filename, False )
        dprint( 'BeginUpdateResource() -> 0x%x - %s' % (h, ctypes.FormatError()) )

        language = 0
        version_id = 1
        rc = UpdateResource( h, RT_VERSION, version_id, language, version_resource, len(version_resource) )
        dprint( 'UpdateResource() -> 0x%x - %s' % (rc, ctypes.FormatError()) )

        rc = EndUpdateResource( h, False );
        dprint( 'EndUpdateResource() => 0x%x' % (rc,) )

        return 0

    def __encodeVersionInfo( self ):
        # pack in reverse order so that sizes are available

        # the tail is copied as is

        # string properties block
        all_string_properties = []

        for name, value in self.all_properties.items():
            sz_name = packSzChar( name )
            if value == '':
                sz_value = b''

            else:
                sz_value = packSzChar( value )

            offset = len(struct_StringHeader) + len(sz_name)

            dprint( 'zzz name %r value %r offset %d' % (name, value, offset) )

            if offset%4 != 0:
                offset += 2
                sz_name += b'\0\0'

            offset += len(sz_value)

            string_header = struct_StringHeader.packer()
            string_header.wLength = offset # qaq
            string_header.wValueLength = len(sz_value)//2
            string_header.wType = 1

            dprint( 'zzz name %r value %r size %d' % (name, value, offset) )

            if offset%4 != 0:
                offset += 2
                sz_value += b'\0\0'

            buf_property = string_header.pack() + sz_name + sz_value
            all_string_properties.append( buf_property )

        buf_all_string_properties = b''.join( all_string_properties )

        # Prefix with an encoding stringheader
        buf_sz_encoding = packSzChar( self.encoding )
        encoding_header = struct_StringHeader.packer()
        encoding_header.wLength = len(encoding_header) + len(buf_all_string_properties) + len(buf_sz_encoding)
        encoding_header.wValueLength = 0
        encoding_header.wType = 1

        buf_properties = encoding_header.pack() + buf_sz_encoding + buf_all_string_properties

        self.pack_string_file_info.wLength = len(self.pack_string_file_info) + len(buf_properties)
        buf_string_properties = self.pack_string_file_info.pack() + buf_properties

        buf_fixed = self.pack_fixed.pack()

        buf_version_info = buf_fixed + buf_string_properties + self.packed_tail

        # VersionInfoHeader
        header = struct_VersionInfoHeader.packer( self.header )
        header.wLength = len(header) + len(buf_version_info)
        buf_header = header.pack()

        version_resource = buf_header + buf_version_info
        return version_resource

    def __loadVersionInfo( self ):
        hmodule = LoadLibraryEx( self.exe_filename, None, LOAD_LIBRARY_AS_DATAFILE_EXCLUSIVE )
        dprint( 'hmodule 0x%x' % (hmodule,) )

        hrsrc = FindResource( hmodule, 1, RT_VERSION )
        dprint( '  hrsrc 0x%x' % (hrsrc,) )

        hglobal = LoadResource( hmodule, hrsrc )
        dprint( 'hglobal 0x%x' % (hglobal,) )

        self.res_size = SizeofResource( hmodule, hrsrc )
        dprint( 'res_size %d' % (self.res_size,) )

        data_p = LockResource( hglobal )
        dprint( ' data_p 0x%x' % (data_p,) )

        buf = ctypes.create_string_buffer( self.res_size )
        ctypes.memmove( buf, data_p, self.res_size )
        data = buf.raw

        # Free handle
        rc = FreeLibrary( hmodule )
        assert rc, 'RC 0x%x' % (rc,)

        self.__decodeVersionInfo( data )

    def __decodeVersionInfo( self, data ):
        dprint( 'VersionInfoResource.__decodeVersionInfo()' )
        hexDump( data, dprint )

        offset = 0
        size = len(struct_VersionInfoHeader)
        assert offset%4 == 0
        self.header = struct_VersionInfoHeader.unpack( data[offset:offset+size] )
        self.header.dump( dprint, 'at offset 0x%x' % (offset,) )

        offset += size
        size = len(struct_VersionInfoFixedFileInfo)
        assert offset%4 == 0
        self.fixed = struct_VersionInfoFixedFileInfo.unpack( data[offset:offset+size] )
        self.fixed.dump( dprint, 'at offset 0x%x' % (offset,) )

        assert offset%4 == 0

        offset += size
        size = len(struct_ChildOfVersionInfoHeader)
        assert offset%4 == 0
        self.child = struct_ChildOfVersionInfoHeader.unpack( data[offset:offset+size] )
        self.child.dump( dprint, 'at offset 0x%x' % (offset,) )

        self.all_properties = {}

        if self.child.wType == 1:
            size = len(struct_StringFileInfo)
            assert offset%4 == 0
            self.string_file_info = struct_StringFileInfo.unpack( data[offset:offset+size] )
            self.string_file_info.dump( dprint, 'at offset 0x%x' % (offset,) )
            dprint( 'szKey %d "%s"' % extractSzChar( data, offset+6 ) )

            end = offset + self.string_file_info.wLength

            offset += size
            assert offset%4 == 0

            size = len(struct_StringHeader)
            assert offset%4 == 0
            string_encoding_header = struct_StringHeader.unpack( data[offset:offset+size] )
            dprint( 'qqq' )
            string_encoding_header.dump( dprint, 'at offset 0x%x' % (offset,) )

            offset += size
            offset, self.encoding = extractSzChar( data, offset )

            while offset < end:
                if offset%4 != 0:
                    offset += 2

                size = len(struct_StringHeader)
                assert offset%4 == 0
                string_header = struct_StringHeader.unpack( data[offset:offset+size] )
                string_header.dump( dprint, 'at offset 0x%x' % (offset,) )

                offset += size
                offset, key = extractSzChar( data, offset )
                if offset%4 != 0:
                    offset += 2

                if string_header.wValueLength == 0:
                    # There is not a string to extract
                    value = ''

                else:
                    offset, value = extractSzChar( data, offset )

                self.all_properties[ key ] = value
                dprint( 'key %d:%r value %d:%r' % (len(key), key, len(value), value) )

            dprint( 'At end: end=%d offset=%d' % (end, offset) )

            if offset%4 != 0:
                offset += 2

            if offset < len(data):
                size = len(struct_ChildOfVersionInfoHeader)
                assert offset%4 == 0
                child = struct_ChildOfVersionInfoHeader.unpack( data[offset:offset+size] )
                child.dump( dprint, 'at offset 0x%x' % (offset,) )

        # the tail bytes stay as is when updating the version resource
        self.packed_tail = data[offset:]

        if( offset < len(data)
        and child.wType == 1 ):
            size = len(struct_VarFileInfo)
            assert offset%4 == 0
            var_file_info = struct_VarFileInfo.unpack( data[offset:offset+size] )
            var_file_info.dump( dprint, 'at offset 0x%x' % (offset,) )

            end = offset + var_file_info.wLength

            offset += size
            assert offset%4 == 0

            while offset < end:
                if offset%4 != 0:
                    offset += 2

                size = len(struct_Var)
                assert offset%4 == 0
                var = struct_Var.unpack( data[offset:offset+size] )
                var.dump( dprint, 'at offset 0x%x' % (offset,) )

                var_end = offset + var.wLength
                offset += size
                if offset % 4 != 0:
                    offset += 2

                while offset < var_end:
                    size = len(struct_VarValue)
                    self.translation = struct_VarValue.unpack( data[offset:offset+size] )
                    self.translation.dump( dprint, 'at offset 0x%x' % (offset,) )
                    offset += size

def hexDump( buffer, write ):
    all_parts = []
    all_chars = []
    for index, byte in enumerate( buffer ):
        if index%16 == 0:
            all_parts.append( '%8.8x' % (index,) )
        all_parts.append( '%2.2x' % (byte, ) )
        if ord(' ') <= byte < 0x7f:
            all_chars.append( chr(byte) )
        else:
            all_chars.append( '.' )

        if index%16 == 15:
            all_parts.reverse()
            all_parts.append( ' %s' % (''.join( all_chars),) )
            write( ' '.join( all_parts ) )
            all_parts = []
            all_chars = []

    if index%16 != 15:
        while index%16 != 15:
            all_parts.append( '  ' )
            index += 1

        all_parts.reverse()
        all_parts.append( ' %s' % (''.join( all_chars),) )
        write( ' '.join( all_parts ) )

def extractSzChar( data, offset ):
    all_chars = []
    while True:
        low = data[offset]
        hi = data[offset+1]
        wchar = (hi<<8) + low
        offset += 2
        if wchar == 0:
            break
        all_chars.append( chr(wchar) )

    return offset, ''.join( all_chars )

def packSzChar( chars ):
    all_bytes = []
    for ch in chars:
        ch = ord(ch)
        lo = ch & 0xff
        hi = (ch>>8) & 0xff

        all_bytes.append( lo )
        all_bytes.append( hi )

    all_bytes.append( 0 )
    all_bytes.append( 0 )

    return bytes( all_bytes )

def updateVersionInfoInExe( exe_filename, version, description ):
    h = BeginUpdateResource( exe_filename, False )
    language = 0

    rc = UpdateResource( h, RT_VERSION, 1, language, data, len(data) )
    print( rc, ctypes.FormatError() )

    rc = EndUpdateResource( h, False );
    print( rc, ctypes.FormatError() )
    return 0

if __name__ == '__main__':
    sys.exit( main( sys.argv ) )



# copied from https://gist.github.com/mkropat/7550097 knownpaths.py has MIT license

import ctypes, sys
from ctypes import windll, wintypes
from uuid import UUID

class GUID(ctypes.Structure):   # [1]
    _fields_ = [
        ("Data1", wintypes.DWORD),
        ("Data2", wintypes.WORD),
        ("Data3", wintypes.WORD),
        ("Data4", wintypes.BYTE * 8)
    ]

    def __init__(self, uuid_):
        ctypes.Structure.__init__(self)
        self.Data1, self.Data2, self.Data3, self.Data4[0], self.Data4[1], rest = uuid_.fields
        for i in range(2, 8):
            self.Data4[i] = rest>>(8 - i - 1)*8 & 0xff

class FOLDERID:     # [2]
    AccountPictures         = UUID('{008ca0b1-55b4-4c56-b8a8-4de4b299d3be}')
    AdminTools              = UUID('{724EF170-A42D-4FEF-9F26-B60E846FBA4F}')
    ApplicationShortcuts    = UUID('{A3918781-E5F2-4890-B3D9-A7E54332328C}')
    CameraRoll              = UUID('{AB5FB87B-7CE2-4F83-915D-550846C9537B}')
    CDBurning               = UUID('{9E52AB10-F80D-49DF-ACB8-4330F5687855}')
    CommonAdminTools        = UUID('{D0384E7D-BAC3-4797-8F14-CBA229B392B5}')
    CommonOEMLinks          = UUID('{C1BAE2D0-10DF-4334-BEDD-7AA20B227A9D}')
    CommonPrograms          = UUID('{0139D44E-6AFE-49F2-8690-3DAFCAE6FFB8}')
    CommonStartMenu         = UUID('{A4115719-D62E-491D-AA7C-E74B8BE3B067}')
    CommonStartup           = UUID('{82A5EA35-D9CD-47C5-9629-E15D2F714E6E}')
    CommonTemplates         = UUID('{B94237E7-57AC-4347-9151-B08C6C32D1F7}')
    Contacts                = UUID('{56784854-C6CB-462b-8169-88E350ACB882}')
    Cookies                 = UUID('{2B0F765D-C0E9-4171-908E-08A611B84FF6}')
    Desktop                 = UUID('{B4BFCC3A-DB2C-424C-B029-7FE99A87C641}')
    DeviceMetadataStore     = UUID('{5CE4A5E9-E4EB-479D-B89F-130C02886155}')
    Documents               = UUID('{FDD39AD0-238F-46AF-ADB4-6C85480369C7}')
    DocumentsLibrary        = UUID('{7B0DB17D-9CD2-4A93-9733-46CC89022E7C}')
    Downloads               = UUID('{374DE290-123F-4565-9164-39C4925E467B}')
    Favorites               = UUID('{1777F761-68AD-4D8A-87BD-30B759FA33DD}')
    Fonts                   = UUID('{FD228CB7-AE11-4AE3-864C-16F3910AB8FE}')
    GameTasks               = UUID('{054FAE61-4DD8-4787-80B6-090220C4B700}')
    History                 = UUID('{D9DC8A3B-B784-432E-A781-5A1130A75963}')
    ImplicitAppShortcuts    = UUID('{BCB5256F-79F6-4CEE-B725-DC34E402FD46}')
    InternetCache           = UUID('{352481E8-33BE-4251-BA85-6007CAEDCF9D}')
    Libraries               = UUID('{1B3EA5DC-B587-4786-B4EF-BD1DC332AEAE}')
    Links                   = UUID('{bfb9d5e0-c6a9-404c-b2b2-ae6db6af4968}')
    LocalAppData            = UUID('{F1B32785-6FBA-4FCF-9D55-7B8E7F157091}')
    LocalAppDataLow         = UUID('{A520A1A4-1780-4FF6-BD18-167343C5AF16}')
    LocalizedResourcesDir   = UUID('{2A00375E-224C-49DE-B8D1-440DF7EF3DDC}')
    Music                   = UUID('{4BD8D571-6D19-48D3-BE97-422220080E43}')
    MusicLibrary            = UUID('{2112AB0A-C86A-4FFE-A368-0DE96E47012E}')
    NetHood                 = UUID('{C5ABBF53-E17F-4121-8900-86626FC2C973}')
    OriginalImages          = UUID('{2C36C0AA-5812-4b87-BFD0-4CD0DFB19B39}')
    PhotoAlbums             = UUID('{69D2CF90-FC33-4FB7-9A0C-EBB0F0FCB43C}')
    PicturesLibrary         = UUID('{A990AE9F-A03B-4E80-94BC-9912D7504104}')
    Pictures                = UUID('{33E28130-4E1E-4676-835A-98395C3BC3BB}')
    Playlists               = UUID('{DE92C1C7-837F-4F69-A3BB-86E631204A23}')
    PrintHood               = UUID('{9274BD8D-CFD1-41C3-B35E-B13F55A758F4}')
    Profile                 = UUID('{5E6C858F-0E22-4760-9AFE-EA3317B67173}')
    ProgramData             = UUID('{62AB5D82-FDC1-4DC3-A9DD-070D1D495D97}')
    ProgramFiles            = UUID('{905e63b6-c1bf-494e-b29c-65b732d3d21a}')
    ProgramFilesX64         = UUID('{6D809377-6AF0-444b-8957-A3773F02200E}')
    ProgramFilesX86         = UUID('{7C5A40EF-A0FB-4BFC-874A-C0F2E0B9FA8E}')
    ProgramFilesCommon      = UUID('{F7F1ED05-9F6D-47A2-AAAE-29D317C6F066}')
    ProgramFilesCommonX64   = UUID('{6365D5A7-0F0D-45E5-87F6-0DA56B6A4F7D}')
    ProgramFilesCommonX86   = UUID('{DE974D24-D9C6-4D3E-BF91-F4455120B917}')
    Programs                = UUID('{A77F5D77-2E2B-44C3-A6A2-ABA601054A51}')
    Public                  = UUID('{DFDF76A2-C82A-4D63-906A-5644AC457385}')
    PublicDesktop           = UUID('{C4AA340D-F20F-4863-AFEF-F87EF2E6BA25}')
    PublicDocuments         = UUID('{ED4824AF-DCE4-45A8-81E2-FC7965083634}')
    PublicDownloads         = UUID('{3D644C9B-1FB8-4f30-9B45-F670235F79C0}')
    PublicGameTasks         = UUID('{DEBF2536-E1A8-4c59-B6A2-414586476AEA}')
    PublicLibraries         = UUID('{48DAF80B-E6CF-4F4E-B800-0E69D84EE384}')
    PublicMusic             = UUID('{3214FAB5-9757-4298-BB61-92A9DEAA44FF}')
    PublicPictures          = UUID('{B6EBFB86-6907-413C-9AF7-4FC2ABF07CC5}')
    PublicRingtones         = UUID('{E555AB60-153B-4D17-9F04-A5FE99FC15EC}')
    PublicUserTiles         = UUID('{0482af6c-08f1-4c34-8c90-e17ec98b1e17}')
    PublicVideos            = UUID('{2400183A-6185-49FB-A2D8-4A392A602BA3}')
    QuickLaunch             = UUID('{52a4f021-7b75-48a9-9f6b-4b87a210bc8f}')
    Recent                  = UUID('{AE50C081-EBD2-438A-8655-8A092E34987A}')
    RecordedTVLibrary       = UUID('{1A6FDBA2-F42D-4358-A798-B74D745926C5}')
    ResourceDir             = UUID('{8AD10C31-2ADB-4296-A8F7-E4701232C972}')
    Ringtones               = UUID('{C870044B-F49E-4126-A9C3-B52A1FF411E8}')
    RoamingAppData          = UUID('{3EB685DB-65F9-4CF6-A03A-E3EF65729F3D}')
    RoamedTileImages        = UUID('{AAA8D5A5-F1D6-4259-BAA8-78E7EF60835E}')
    RoamingTiles            = UUID('{00BCFC5A-ED94-4e48-96A1-3F6217F21990}')
    SampleMusic             = UUID('{B250C668-F57D-4EE1-A63C-290EE7D1AA1F}')
    SamplePictures          = UUID('{C4900540-2379-4C75-844B-64E6FAF8716B}')
    SamplePlaylists         = UUID('{15CA69B3-30EE-49C1-ACE1-6B5EC372AFB5}')
    SampleVideos            = UUID('{859EAD94-2E85-48AD-A71A-0969CB56A6CD}')
    SavedGames              = UUID('{4C5C32FF-BB9D-43b0-B5B4-2D72E54EAAA4}')
    SavedSearches           = UUID('{7d1d3a04-debb-4115-95cf-2f29da2920da}')
    Screenshots             = UUID('{b7bede81-df94-4682-a7d8-57a52620b86f}')
    SearchHistory           = UUID('{0D4C3DB6-03A3-462F-A0E6-08924C41B5D4}')
    SearchTemplates         = UUID('{7E636BFE-DFA9-4D5E-B456-D7B39851D8A9}')
    SendTo                  = UUID('{8983036C-27C0-404B-8F08-102D10DCFD74}')
    SidebarDefaultParts     = UUID('{7B396E54-9EC5-4300-BE0A-2482EBAE1A26}')
    SidebarParts            = UUID('{A75D362E-50FC-4fb7-AC2C-A8BEAA314493}')
    SkyDrive                = UUID('{A52BBA46-E9E1-435f-B3D9-28DAA648C0F6}')
    SkyDriveCameraRoll      = UUID('{767E6811-49CB-4273-87C2-20F355E1085B}')
    SkyDriveDocuments       = UUID('{24D89E24-2F19-4534-9DDE-6A6671FBB8FE}')
    SkyDrivePictures        = UUID('{339719B5-8C47-4894-94C2-D8F77ADD44A6}')
    StartMenu               = UUID('{625B53C3-AB48-4EC1-BA1F-A1EF4146FC19}')
    Startup                 = UUID('{B97D20BB-F46A-4C97-BA10-5E3608430854}')
    System                  = UUID('{1AC14E77-02E7-4E5D-B744-2EB1AE5198B7}')
    SystemX86               = UUID('{D65231B0-B2F1-4857-A4CE-A8E7C6EA7D27}')
    Templates               = UUID('{A63293E8-664E-48DB-A079-DF759E0509F7}')
    UserPinned              = UUID('{9E3995AB-1F9C-4F13-B827-48B24B6C7174}')
    UserProfiles            = UUID('{0762D272-C50A-4BB0-A382-697DCD729B80}')
    UserProgramFiles        = UUID('{5CD7AEE2-2219-4A67-B85D-6C9CE15660CB}')
    UserProgramFilesCommon  = UUID('{BCBD3057-CA5C-4622-B42D-BC56DB0AE516}')
    Videos                  = UUID('{18989B1D-99B5-455B-841C-AB7C74E4DDFC}')
    VideosLibrary           = UUID('{491E922F-5643-4AF4-A7EB-4E7A138D8174}')
    Windows                 = UUID('{F38BF404-1D43-42F2-9305-67DE0B28FC23}')

class UserHandle:   # [3]
    current = wintypes.HANDLE(0)
    common  = wintypes.HANDLE(-1)

_CoTaskMemFree = windll.ole32.CoTaskMemFree     # [4]
_CoTaskMemFree.restype= None
_CoTaskMemFree.argtypes = [ctypes.c_void_p]

_SHGetKnownFolderPath = windll.shell32.SHGetKnownFolderPath     # [5] [3]
_SHGetKnownFolderPath.argtypes = [
    ctypes.POINTER(GUID), wintypes.DWORD, wintypes.HANDLE, ctypes.POINTER(ctypes.c_wchar_p)
]

class PathNotFoundException(Exception): pass

def get_path(folderid, user_handle=UserHandle.common):
    fid = GUID(folderid)
    pPath = ctypes.c_wchar_p()
    S_OK = 0
    if _SHGetKnownFolderPath(ctypes.byref(fid), 0, user_handle, ctypes.byref(pPath)) != S_OK:
        raise PathNotFoundException()
    path = pPath.value
    _CoTaskMemFree(pPath)
    return path

if __name__ == '__main__':
    if len(sys.argv) < 2 or sys.argv[1] in ['-?', '/?']:
        print('python knownpaths.py FOLDERID {current|common}')
        sys.exit(0)

    try:
        folderid = getattr(FOLDERID, sys.argv[1])
    except AttributeError:
        print('Unknown folder id "%s"' % sys.argv[1], file=sys.stderr)
        sys.exit(1)

    try:
        if len(sys.argv) == 2:
            print(get_path(folderid))
        else:
            print(get_path(folderid, getattr(UserHandle, sys.argv[2])))
    except PathNotFoundException:
        print('Folder not found "%s"' % ' '.join(sys.argv[1:]), file=sys.stderr)
        sys.exit(1)

# [1] http://msdn.microsoft.com/en-us/library/windows/desktop/aa373931.aspx
# [2] http://msdn.microsoft.com/en-us/library/windows/desktop/dd378457.aspx
# [3] http://msdn.microsoft.com/en-us/library/windows/desktop/bb762188.aspx
# [4] http://msdn.microsoft.com/en-us/library/windows/desktop/ms680722.aspx
# [5] http://www.themacaque.com/?p=954



#
#   __init__.py
#
VERSION = '1.4.0'

__commands = set( ['build', 'flags'] )

def dispatchCommand( argv ):
    if len(argv) < 2 or argv[1] not in __commands:
        from . import win_app_package_builder
        win_app_package_builder.AppPackage().usage()

        from . import win_app_package_exe_config
        win_app_package_exe_config.usage()

        return 1

    if argv[1] == 'build':
        from . import win_app_package_builder
        return win_app_package_builder.AppPackage().buildCommand( argv )

    elif argv[1] == 'flags':
        from . import win_app_package_exe_config
        return win_app_package_exe_config.flagsCommand( argv )

    else:
        print( 'Error: Unknown command %r' % (argv[1],) )
        return 1



#
#   run_win_app.py
#
#   Inspired by Python's runpy.py
#
import sys
import types

def run_win_app( path_name ):
    with open( path_name, "rb" ) as f:
        code = compile( f.read(), path_name, 'exec' )

    module = types.ModuleType( '__main__' )
    sys.modules[ '__main__' ] = module
    module.__dict__.update(
        __name__ = '__main__',
        __file__ = path_name,
        __cached__ = None,
        __doc__ = None,
        __loader__ = None,
        __package__ = '',
        __spec__ = None
        )
    exec( code, module.__dict__ )



#!/usr/bin/python3
import namedstruct
import pathlib

dos_magic = b'MZ'
struct_dos_header = namedstruct.namedstruct( 'DOS header', '<'
    '2s:dos_magic '
    'h:lastsize '
    'h:nblocks '
    'h:nreloc '
    'h:hdrsize '
    'h:minalloc '
    'h:maxalloc '
    'H:ss '
    'H:sp '
    'h:checksum '
    'H:ip '
    'H:cs '
    'h:relocpos '
    'h:noverlay '
    '4H:reserved1 '
    'h:oem_id '
    'h:oem_info '
    '10H:reserved2 '
    'L:pe_offset'
    )

pe_magic = b'PE\x00\x00'
struct_pe_header = namedstruct.namedstruct( 'PE Header', '<'
    '4s:pe_magic'
    )

struct_coff_header = namedstruct.namedstruct( 'COFF Header', '<'
    'H:Machine '
    'h:NumberOfSections '
    'l:TimeDateStamp '
    'l:PointerToSymbolTable '
    'l:NumberOfSymbols '
    'h:SizeOfOptionalHeader '
    'h:Characteristics '
    )

struct_pe_opt_signature = namedstruct.namedstruct( 'PE signature', '<'
    'H:signature '
    )
pe32_signature = 0x10b
pe64_signature = 0x20b

struct_pe32_opt_header = namedstruct.namedstruct( 'PE 32 Opt Header', '<'
    'H:signature '
    'B:MajorLinkerVersion '
    'B:MinorLinkerVersion '
    'L:SizeOfCode '
    'L:SizeOfInitializedData '
    'L:SizeOfUninitializedData '
    'L:AddressOfEntryPoint '
    'L:BaseOfCode '
    'L:BaseOfData '
    'L:ImageBase '
    'L:SectionAlignment '
    'L:FileAlignment '
    'H:MajorOSVersion '
    'H:MinorOSVersion '
    'H:MajorImageVersion '
    'H:MinorImageVersion '
    'H:MajorSubsystemVersion '
    'H:MinorSubsystemVersion '
    'L:Reserved '
    'L:SizeOfImage '
    'L:SizeOfHeaders '
    'L:Checksum '
    'H:Subsystem '
    'H:DLLCharacteristics '
    'L:SizeOfStackReserve '
    'L:SizeOfStackCommit '
    'L:SizeOfHeapReserve '
    'L:SizeOfHeapCommit '
    'L:LoaderFlags '
    'L:NumberOfRvaAndSizes'
    #data_directory DataDirectory[16];
    )

struct_pe64_opt_header = namedstruct.namedstruct( 'PE 64 Opt Header', '<'
    'H:signature '
    'B:MajorLinkerVersion '
    'B:MinorLinkerVersion '
    'L:SizeOfCode '
    'L:SizeOfInitializedData '
    'L:SizeOfUninitializedData '
    'L:AddressOfEntryPoint '
    'L:BaseOfCode '
    'Q:ImageBase '
    'L:SectionAlignment '
    'L:FileAlignment '
    'H:MajorOSVersion '
    'H:MinorOSVersion '
    'H:MajorImageVersion '
    'H:MinorImageVersion '
    'H:MajorSubsystemVersion '
    'H:MinorSubsystemVersion '
    'L:Reserved '
    'L:SizeOfImage '
    'L:SizeOfHeaders '
    'L:Checksum '
    'H:Subsystem '
    'H:DLLCharacteristics '
    'Q:SizeOfStackReserve '
    'Q:SizeOfStackCommit '
    'Q:SizeOfHeapReserve '
    'Q:SizeOfHeapCommit '
    'L:LoaderFlags '
    'L:NumberOfRvaAndSizes'
    #data_directory DataDirectory[16];
    )

IMAGE_DIRECTORY_ENTRY_IMPORT = 1
struct_pe_data_dir = namedstruct.namedstruct( 'PE Data directory', '<'
    'L:VirtualAddress '
    'L:Size'
    )

struct_coff_section = namedstruct.namedstruct( 'COFF Section', '<'
    '8s:Name '
    'L:VirtualSize '
    'L:VirtualAddress '
    'L:SizeOfRawData '
    'L:PointerToRawData '
    'L:PointerToRelocations '
    'L:PointerToLinenumbers '
    'H:NumberOfRelocations '
    'H:NumberOfLinenumbers '
    'L:Characteristics '
    )

struct_import_dirent = namedstruct.namedstruct( 'Import directory entry', '<'
    'L:ImportLookupTable '
    'L:TimeDateStamp '
    'L:ForwarderChain '
    'L:NameRVA '
    'L:ImportAddressTable'
    )


def getPeImportDlls( log, filename ):
    with filename.open( 'rb' ) as f:
        pe_image = f.read()

    dos_header = struct_dos_header.unpack( pe_image[0:len(struct_dos_header)] )
    log.debug( 'dos_header.dos_magic %r .pe_offset %r' % (dos_header.dos_magic, dos_header.pe_offset) )
    if dos_header.dos_magic != dos_magic:
        log.error( 'Not a PE image, bad dos magic %r in %s' % (dos_header.dos_magic, filename) )
        dos_header.dump( log.error )
        return set()

    pe_start = dos_header.pe_offset

    pe_header= struct_pe_header.unpack( pe_image[pe_start:pe_start+len(struct_pe_header)] )
    log.debug( 'pe_header.pe_magic %r' % (pe_header.pe_magic,) )
    if pe_header.pe_magic != pe_magic:
        log.error( 'Not a PE image, bad PE magic %r in %s' % (pe_header.pe_magic, filename) )
        log.error( 'DOS header' )
        dos_header.dump( log.error )
        log.error( 'PE header' )
        pe_header.dump( log.error )
        return set()

    coff_start = pe_start + len(struct_pe_header)

    coff_header = struct_coff_header.unpack( pe_image[coff_start:coff_start+len(struct_coff_header)] )
    log.debug( 'COFF header @ 0x%8.8x' % (coff_start,) )
    coff_header.dump( log.debug )

    pe_opt_start = coff_start + len(struct_coff_header)

    pe_opt_signature = struct_pe_opt_signature.unpack( pe_image[pe_opt_start:pe_opt_start+len(struct_pe_opt_signature)] )
    if pe_opt_signature.signature ==  pe32_signature:
        struct_pe_opt_header = struct_pe32_opt_header

    elif pe_opt_signature.signature == pe64_signature:
        struct_pe_opt_header = struct_pe64_opt_header

    else:
        log.error( 'Unknown PE optional header signature of 0x4.4x' % (pe_opt_signature.signature,) )
        return set()

    pe_opt_header = struct_pe_opt_header.unpack( pe_image[pe_opt_start:pe_opt_start+len(struct_pe_opt_header)] )

    log.debug( 'PE Opt header @ 0x%8.8x' % (pe_opt_start,) )
    pe_opt_header.dump( log.debug )

    # Decode the Data Directories
    data_dir_0_start = pe_opt_start + len(struct_pe_opt_header)

    all_data_dir = []

    for index in range( pe_opt_header.NumberOfRvaAndSizes ):
        data_dir_start = data_dir_0_start + len(struct_pe_data_dir)*index

        pe_data_dir= struct_pe_data_dir.unpack( pe_image[data_dir_start:data_dir_start+len(struct_pe_data_dir)] )

        log.debug( 'Data Dir %d @ 0x%8.8x' % (index, data_dir_start) )
        pe_data_dir.dump( log.debug )

        all_data_dir.append( pe_data_dir )

    # Decode the COFF Sections
    section_0_start = data_dir_0_start + pe_opt_header.NumberOfRvaAndSizes * len(struct_pe_data_dir)

    all_coff_sections = []
    for index in range( coff_header.NumberOfSections ):
        section_start = section_0_start + index * len(struct_coff_section)
        coff_section = struct_coff_section.unpack( pe_image[section_start:section_start+len(struct_coff_section)] )

        log.debug( 'Section %d @ 0x%8.8x' % (index, section_start) )
        coff_section.dump( log.debug )

        all_coff_sections.append( coff_section )

    # convert a RVA to the offset in the disk image
    def rvaToDisk( addr ):
        for section in all_coff_sections:
            if section.VirtualAddress <= addr < (section.VirtualAddress + section.VirtualSize):
                offset = addr - section.VirtualAddress
                return section.PointerToRawData + offset

        raise ValueError( 'RVA 0x%8.8x is not within any COFF section in file %s' % (addr, filename) )

    end = section_0_start + coff_header.NumberOfSections * len(struct_coff_section)

    log.debug( 'Decode @ 0x%8.8x' % (end,) )

    import_data_dir = all_data_dir[ IMAGE_DIRECTORY_ENTRY_IMPORT ]
    log.debug( 'import_data_dir %r' % (import_data_dir,) )
    import_data_dir.dump( log.debug )

    if import_data_dir.Size == 0:
        log.debug( 'Empty import list' )
        return set()

    start_import_data_dir = rvaToDisk( import_data_dir.VirtualAddress )

    log.debug( 'import data 0x%8.8x' % (start_import_data_dir,) )

    def unpackString( addr ):
        start = addr
        for i in range( 256 ):
            if pe_image[start+i] == 0:
                break

        return pe_image[start:start+i]

    all_dlls = set()

    while True:
        import_dirent = struct_import_dirent.unpack( pe_image[start_import_data_dir:start_import_data_dir+len(struct_import_dirent)] )
        log.debug( 'Import dirent @ 0x%8.8x' % (start_import_data_dir,) )
        import_dirent.dump( log.debug )
        if import_dirent.ImportLookupTable == 0:
            break

        start_name = rvaToDisk( import_dirent.NameRVA )
        name = unpackString( start_name )
        log.debug( 'Name @ 0x%8.8x %r' % (start_name, name) )

        all_dlls.add( pathlib.Path( name.decode('ASCII' ) ) )

        start_import_data_dir += len(struct_import_dirent)

    log.debug( 'OK for %s' % (filename,) )

    return all_dlls

if __name__ == '__main__':
    class log:
        def info( self, m ):
            print( 'Info: %s' % (m,) )

        def debug( self, m ):
            print( 'Debug: %s' % (m,) )

        def verbose( self, m ):
            print( 'Verbose: %s' % (m,) )

        def error( self, m ):
            print( 'Error: %s' % (m,) )

    print( getPeImportDlls( log(), pathlib.Path( r'C:\python36.win64\python3.dll' ) ) )



#/usr/bin/python3
#
#   win_app_package_builder.py
#
import sys
import os
import pathlib
import uuid
import modulefinder
import importlib
import colour_text

from . import win_app_package_win_pe_info
from . import win_app_package_exe_config
from . import win_known_paths

class AppPackageError(Exception):
    pass

class AppPackage:
    APP_TYPE_CLI = 1
    APP_TYPE_GUI = 2

    resource_folder = pathlib.Path( 'PyWinAppRes' )
    library_folder = resource_folder / 'lib'

    all_modules_allowed_to_be_missing = set( [
        'sets',
        '_dummy_threading',
        '_frozen_importlib',
        '_frozen_importlib_external',
        '_posixsubprocess',
        '_scproxy',
        '_winreg',
        'ce',
        'grp',
        'java.lang',
        'org.python.core',
        'os.path',
        'posix',
        'pwd',
        'readline',
        'termios',
        'vms_lib',

        # condition import for python2 compat which can be ignored
        '_perf',
        'cStringIO',
        'ordereddict',
        'sha',
        'UserDict',

        # from python2 support for pkg_resources via pytz
        '_sysconfigdata',
        'multiprocessing.get_context',
        'multiprocessing.BufferTooShort',
        'dummy.Process',
        'multiprocessing.AuthenticationError',
        'urlparse',
        'packaging.specifiers',
        'multiprocessing.set_start_method',
        'multiprocessing.TimeoutError',
        'packaging.version',
        'multiprocessing.get_start_method',

        # git - module can be ignored
        'git.index.IndexFile',
        'git.objects.Object',
        'git.objects.RootModule',
        'git.refs.RemoteReference',
        'git.refs.SymbolicReference',
        'git.refs.TagReference',
        'git.repo.Repo',
        'refs.RemoteReference',
        'refs.SymbolicReference',
        'refs.TagReference',
        'smmap.SlidingWindowMapBuffer',
        'smmap.SlidingWindowMapManager',
        'smmap.StaticWindowMapManager',
        ] )

    all_imported_modules_to_exclude = set( [
        'ctypes',
        'ctypes._endian',
        'ctypes.util',
        'ctypes.wintypes',
        'importlib',
        'importlib._bootstrap',
        'importlib._bootstrap_external',
        'importlib.abc',
        'importlib.machinery',
        'importlib.util',
        'modulefinder',
        'uuid',
        'win_app_packager',
        'win_app_packager.win_app_package_builder',
        'win_app_packager.win_app_package_exe_config',
        'win_app_packager.win_app_package_win_pe_info',
        ] )

    def __init__( self ):
        self.ct = colour_text.ColourText()
        self.ct.initTerminal()

        # options
        self.enable_debug = False
        self.enable_verbose = False
        self.enable_merge = False

        self.enable_bootstrap_debug = False

        self.app_type = self.APP_TYPE_CLI
        self.app_name = None
        self.app_icon = None
        self.app_version = (0, 0, 0, 0)
        self.app_install_key = ''
        self.app_install_value = ''

        self.modules_allowed_to_be_missing_filename = None

        self.main_program = None
        self.package_folder = None

        # need to check for both paths
        # the cases covered are:
        # win32 app on windows 32 bit system
        # win32 app on windows 64 bit system
        # win64 app on windows 64 bit system
        self.__windows_system_folders = (pathlib.Path( win_known_paths.get_path( win_known_paths.FOLDERID.System ) )
                                        ,pathlib.Path( win_known_paths.get_path( win_known_paths.FOLDERID.SystemX86 ) ))

        # package contents
        self.__all_library_files = set()
        self.__all_dlls = set()

        self.__all_found_dlls = set()

        # how to find app packager resources
        self.win_app_packager_folder = pathlib.Path( sys.argv[0] ).parent

    def debug( self, msg ):
        if self.enable_debug:
            print( 'Debug: %s' % (msg,) )

    def info( self, msg ):
        print( self.ct('<>info Info:<> %s') % (msg,) )

    def verbose( self, msg ):
        if self.enable_verbose:
            print( self.ct('<>info Info:<> %s') % (msg,) )

    def error( self, msg ):
        print( self.ct('<>error Error: %s<>') % (msg,) )

    def warning( self, msg ):
        print( self.ct('<>em Warn:<> %s') % (msg,) )

    def usage( self ):
################################################################################
        print(
'''python3 -m win_app_packager build <main-script> <package-folder> [<options>...]
  main-script
    - python main module
  package-folder
    - folder to create package into

  Where <options> are:
    --console
    --cli
        build a windows console progam (the default).
    --gui
        build a windows gui program.
    --name
        name the program (defaults to the <main-script> name).
    --version <version>
        Set the version of the .EXE to be <version>.
        e.g 1.0.2.5
    --install-key <key>
    --install-value <value>
        The install path of the package can be read
        from the windows registry from key HKLM:<key> value <value>
        otherwise the install path is assumed to be the same folder
        that the .EXE files is in.
    --modules-allowed-to-be-missing-file <file-name>
        Add all the modules listed in the file <file-name> to the allowed
        to be missing list. Blank lines and lines starting with a '#' as ignored.
    --merge
        Do not clean out the <package-folder> before building the package.
        Useful for putting multiple programs into one package.
    --verbose
        Output extra information about the build process.
    --debug
        Developer option. Output lots of details about the build process.
    --bootstrap-debug
        Developer option. Copy PDF files and setup a Microsoft Visual
        Studio solution (.sln) file suitable for running the bootstrap
        under the debugger.
'''
)
        return 1

    def parseArgs( self, argv ):
        all_positional_args = []
        args = iter( argv )
        next(args)

        for arg in args:
            if arg.startswith( '--' ):
                if arg == '--debug':
                    self.enable_debug = True

                elif arg == '--bootstrap-debug':
                    self.enable_bootstrap_debug = True

                elif arg == '--verbose':
                    self.enable_verbose = True

                elif arg in ('--console', '--cli'):
                    self.app_type = self.APP_TYPE_CLI

                elif arg == '--gui':
                    self.app_type = self.APP_TYPE_GUI

                elif arg == '--name':
                    self.app_name = next(args)

                elif arg == '--install-key':
                    self.app_install_key = next(args)

                elif arg == '--install-value':
                    self.app_install_value = next(args)

                elif arg == '--icon':
                    self.app_icon = next(args)

                elif arg == '--version':
                    # expecting upto 4 int seperated by "."
                    try:
                        version = next(args)
                        if version.strip() == '':
                            raise AppPackageError( '--version - value must not be empty' )

                        int_version = list( int(n) for n in version.split('.') )
                        # pad with 0 to make exactly 4 parts to the version
                        while len(int_version) < 4:
                            int_version.append( 0 )

                        if len(int_version) > 4:
                            raise AppPackageError( '--version %r - only 4 parts allowed' )

                    except StopIteration:
                        raise AppPackageError( '--version - value required' )

                    except ValueError:
                        raise AppPackageError( '--version - invalid value %r' % (version,) )

                    self.app_version = tuple(int_version)

                elif arg == '--merge':
                    self.enable_merge = True

                elif arg == '--modules-allowed-to-be-missing-file':
                    self.modules_allowed_to_be_missing_filename = next(args)

                else:
                    raise AppPackageError( 'Unknown option %r' % (arg,) )

            else:
                all_positional_args.append( arg )

        if( len( all_positional_args ) < 1
        or all_positional_args[0] != 'build' ):
            raise AppPackageError( 'Expecting command name "build"' )

        if len( all_positional_args ) < 3:
            raise AppPackageError( 'build expects two args' )

        self.main_program = all_positional_args[1]
        self.package_folder = pathlib.Path( all_positional_args[2] ).resolve()

        if self.app_name is None:
            self.app_name = self.main_program[:-len('.py')]

        if self.app_install_key != '' and self.app_install_value == '':
            raise AppPackageError( 'require --install-value with --install-key' )

    def buildCommand( self, argv ):
        try:
            self.info( 'App Package Builder' )
            self.parseArgs( argv )

            self.processModulesAllowedToBeMissingFile()

            if self.app_type == self.APP_TYPE_CLI:
                self.info( 'Building CLI App %s into package folder %s' % (self.app_name, self.package_folder) )

            elif self.app_type == self.APP_TYPE_GUI:
                self.info( 'Building GUI App %s into package folder %s' % (self.app_name, self.package_folder) )

            else:
                raise AppPackageError( 'Unknown app_type %r' % (self.app_type,) )


            #
            #   Look for modules using two methods
            #   1. Import the main program and see what ends up in sys.modules
            #   2. Force "encodings" to be included as python will not start without encodings
            #   3. Use modulefinder to locate imports done at runtime
            #

            # 1. import main program
            main_module = self.main_program
            if main_module.endswith( '.py' ):
                main_module = main_module[:-len('.py')]

            self.info( 'Importing %s' % (main_module,) )
            importlib.import_module( main_module )
            self.info( 'Import complete for %s' % (main_module,) )

            # save the list of modules imported
            all_imported_module_names = list( sys.modules.keys() )

            # 2. force in encodings
            importlib.import_module( 'encodings' )

            # 3. what can modulefinder locate
            mf = modulefinder.ModuleFinder()
            mf.run_script( self.main_program )
            for name, mod in sorted( mf.modules.items() ):
                self.verbose( 'Module %s: %r' % (name, mod) )

            missing, maybe = mf.any_missing_maybe()
            all_missing = set( missing )
            all_missing_but_needed = all_missing - self.all_modules_allowed_to_be_missing

            for x in all_missing_but_needed:
                self.error( 'Module %s is missing but is required' % (x,) )

            for x in maybe:
                self.warning( 'Module %s maybe missing' % (x,) )

            if len(all_missing_but_needed) > 0:
                return 1

            # find the python DLL
            self.addWinPeFileDependenciesToPackage( pathlib.Path( sys.executable ) )

            for name, mod in sorted( mf.modules.items() ):
                self.processModule( name, mod )

            for name in sorted( all_imported_module_names ):
                if name in self.all_imported_modules_to_exclude:
                    continue

                self.processModule( name, sys.modules[ name ] )

            if not self.enable_merge:
                self.cleanAppPackage()

            self.createAppPackage()

            self.info( 'Completed sucessfully' )
            return 0

        except AppPackageError as e:
            self.error( str(e) )
            return 1

    def processModulesAllowedToBeMissingFile( self ):
        if self.modules_allowed_to_be_missing_filename is None:
            return

        try:
            with open( self.modules_allowed_to_be_missing_filename, 'r', encoding='utf=8' ) as f:
                for line in f:
                    line = line.strip()
                    if line == '' or line.startswith( '#' ):
                        continue

                    self.verbose( 'Adding module "%s" to the allowed to be missing modules list' % (line,) )
                    self.all_modules_allowed_to_be_missing.add( line )

        except FileNotFoundError as e:
            raise AppPackageError( str(e) )

    def processModule( self, name, module ):
        if not hasattr( module, '__file__' ) or module.__file__ is None:
            self.verbose( 'Module %s is builtin - ignoring' % (name,) )
            return

        filename = pathlib.Path( module.__file__ ).resolve()
        self.verbose( 'Module %s type %s filename %s' % (name, filename.suffix, filename) )

        # is this file part of the python installation?
        for root in [sys.base_prefix] + sys.path:
            try:
                root = pathlib.Path( root )
                root = root.resolve()

                # find the suffix relative to a python home
                library_filename_suffix = filename.relative_to( root )

            except FileNotFoundError:
                # root does not exist
                continue

            except ValueError:
                # of filename is not relative to root
                continue

            if filename.match( '*.py' ):
                self.verbose( 'Module %s suffix %s' % (name, library_filename_suffix) )
                self.addPyFileToPackage( filename, library_filename_suffix )
                if filename.name == '__init__.py':
                    # assume that all files in the module need including
                    self.addAllModuleFilesToPackage( filename.parent, library_filename_suffix.parent )

            elif( filename.match( '*.pyd' )
            or    filename.match( '*.dll' ) ):
                self.addWinPeFileToPackage( filename, library_filename_suffix )

            else:
                raise AppPackageError( 'No handler for files of type %s' % (ext,) )

            return

        raise AppPackageError( 'Cannot find required file %s' % (filename,) )

    def addAllModuleFilesToPackage( self, module_dirname, library_dirname_suffix ):
        self.verbose( 'Adding module files to %s from %s' % (library_dirname_suffix, module_dirname) )
        for filename in module_dirname.iterdir():
            if filename.suffix in ('.pyc', ',pyo'):
                continue

            if filename.is_dir():
                self.addAllModuleFilesToPackage( filename, library_dirname_suffix / filename.name )

            else:
                self.addPyFileToPackage( filename, library_dirname_suffix / filename.name )

    def addPyFileToPackage( self, filename, library_filename_suffix ):
        self.verbose( 'Adding source file %s from %s' % (library_filename_suffix, filename) )
        self.__all_library_files.add( PackageFile( self, filename, library_filename_suffix ) )

    def addWinPeFileToPackage( self, filename, library_filename_suffix ):
        self.debug( 'AppPackage.addWinPeFileToPackage( %r, %r )' % (filename, library_filename_suffix) )
        pf = PackageFile( self, filename, library_filename_suffix )
        if pf not in self.__all_dlls:
            self.verbose( 'Adding DLL %s from %s' % (library_filename_suffix, filename) )
            self.__all_dlls.add( pf )

            self.addWinPeFileDependenciesToPackage( filename )

    def addWinPeFileDependenciesToPackage( self, filename ):
        self.debug( 'AppPackage.addWinPeFileDependenciesToPackage( %r )' % (filename,) )
        all_dlls = win_app_package_win_pe_info.getPeImportDlls( self, filename )
        self.verbose( 'Dependancies of DLL %s:' % (filename,) )
        for dll in sorted( all_dlls ):
            self.verbose( '   %s' % (dll,) )

        all_dll_to_be_scanned = []

        for dll in sorted( all_dlls ):
            # only need to scan each DLL once
            if dll in self.__all_found_dlls:
                continue
            self.__all_found_dlls.add( dll )

            dll_path = self.findDll( dll, filename.parent )
            if self.isStandardDll( dll_path ):
                self.verbose( 'No need to package standard DLL %s' % (dll_path,) )
                continue

            all_dll_to_be_scanned.append( (dll_path, self.findPackageSuffixForDll( dll_path )) )

        for dll_path, library_filename_suffix in all_dll_to_be_scanned:
            self.addWinPeFileToPackage( pathlib.Path( dll_path ), library_filename_suffix )

    def findPackageSuffixForDll( self, dll_path ):
        # either its in the prefix location or on the path.
        # use sys.base_prefix so that this works for venv or not venv.
        # put prefix first so that python*.dll and vcruntime*.dll will be found
        # as shipped with the python installation.
        for root in  [sys.base_prefix] + sys.path:
            try:
                root = pathlib.Path( root )
                path = root / dll_path;
                self.debug( 'Looking for %s in %s' % (dll_path, path) )

                path = pathlib.Path( path ).resolve()
                if not path.exists():
                    continue

                suffix = dll_path.relative_to( root )
                self.debug( 'DLL root %s suffix %s' % (root, suffix) )

                return suffix

            except FileNotFoundError:
                self.debug( '-- FileNotFoundError' )
                continue

            except ValueError as e:
                self.debug( '-- ValueError %s' % (e,) )
                continue

        raise AppPackageError( 'Cannot find required DLL %s' % (dll_path,) )

    def findDll( self, dll_name, preferred_location ):
        # look for the DLL in an ordered set of locations
        # starting with the 'preferred_location'
        # and then working along the 'PATH'

        self.verbose( 'Find DLL %s (preferred_location %s)' % (dll_name, preferred_location) )
        for folder in [sys.base_prefix, preferred_location]+os.environ['PATH'].split( os.pathsep ):
            try:
                folder = pathlib.Path( folder ).resolve()

            except FileNotFoundError:
                continue

            except PermissionError:
                continue

            dll_path = folder / dll_name
            if dll_path.exists():
                return dll_path

        return dll_name

    def isStandardDll( self, dll ):
        dll = pathlib.Path( dll )

        if dll.match( 'api-ms-win-*.dll' ):
            return True

        if dll.parent in self.__windows_system_folders:
            return True

        if dll.name == 'dbghelp.dll':
            return True

        return False

    def cleanAppPackage( self ):
        #
        # Clean up the package_folder, but only if
        # the special PyWinAppLib folder is within it.
        #
        # This is to defend against removing files
        # from unintended locations like c:\windows
        #
        if not self.package_folder.exists():
            return

        if not self.package_folder.is_dir():
            raise AppPackageError( '%s is not a directory' )

        all_filenames = list( self.package_folder.iterdir() )
        if len(all_filenames) == 0:
            return

        if (self.package_folder / self.resource_folder) not in all_filenames:
            raise AppPackageError( '%s is not an existing packager folder - refusing to clean up' % (self.package_folder,) )

        self.info( 'Cleaning up %s' % (self.package_folder,) )
        for dirpath, all_dirnames, all_filenames in os.walk( str( self.package_folder ), topdown=False ):
            for filename in all_filenames:
                (pathlib.Path( dirpath ) / filename).unlink()

            for dirname in all_dirnames:
                (pathlib.Path( dirpath ) / dirname).rmdir()

    def createAppPackage( self ):
        #
        #   create the following structure on disk
        #
        #   <package-folder>\<main-program>.exe
        #   <package-folder>\PyWinAppLib\<library-file>
        #

        self.info( 'Creating App Package in %s' % (self.package_folder,) )

        path_library_folder = pathlib.Path( self.package_folder ) / self.library_folder
        path_resource_folder = pathlib.Path( self.package_folder ) / self.resource_folder

        # copy the programs python source code
        for lib_file in sorted( self.__all_library_files ):
            lib_file.copy( self.ct('Copying Library file <>em %s<>'), path_resource_folder )

        # copy the programs DLL dependencies
        for dll_file in sorted( self.__all_dlls ):
            dll_file.copy( self.ct('Copying DLL <>em %s<>'), path_resource_folder )

            if self.enable_bootstrap_debug:
                # include PDB file need to support debugging
                dll_file.copyPdb( path_resource_folder )

        # copy the right boot strap
        if self.app_type == self.APP_TYPE_CLI:
            bootstrap_filename = self.win_app_packager_folder / 'BootStrap' / 'obj' / 'bootstrap-cli.exe'

        elif self.app_type == self.APP_TYPE_GUI:
            bootstrap_filename = self.win_app_packager_folder / 'BootStrap' / 'obj' / 'bootstrap-gui.exe'

        bootstrap_exe = pathlib.Path( '%s.exe' % (self.app_name,) )
        p = PackageFile( self, bootstrap_filename, bootstrap_exe )
        p.copy( self.ct('Package file <>em %s<>'), self.package_folder )

        p = PackageFile(
                self,
                self.win_app_packager_folder / 'run_win_app.py',
                self.library_folder / 'run_win_app.py' )
        p.copy( self.ct('Package file <>em %s<>'), self.package_folder )

        if self.enable_bootstrap_debug:
            # include PDB file need to support debugging
            p.copyPdb( self.package_folder )

            #
            # create the Microsoft Visual Studio 14.0 solution file
            # need to run the boot strap  code under the debugger
            #
            sln_vars = {
                'UUID1':                uuid.uuid4(),
                'UUID2':                uuid.uuid4(),
                'name':                 self.app_name,
                'exefile':              bootstrap_exe,
                'Executable':           self.package_folder.resolve() / bootstrap_exe,
                'StartingDirectory':    self.package_folder.resolve(),
                }
            sln_file = (self.package_folder / bootstrap_exe ).with_suffix( '.sln' )
            sln_file.write_text( self.vc_14_solution_file_template % sln_vars )

        win_app_package_exe_config.configureAppExeBootStrap(
            str( self.package_folder / bootstrap_exe ),
            'python%d%d.dll' % (sys.version_info.major, sys.version_info.minor),
            self.main_program,
            self.app_install_key,
            self.app_install_value,
            )

        if self.app_icon is not None:
            self.info( 'Settings ICON from %s' % (self.app_icon,) )
            win_app_package_exe_config.updateIconInExe(
                str( self.package_folder / bootstrap_exe ),
                self.app_icon
                )

        all_changes = ['FileDescription', self.app_name
                      ,'ProductName', self.app_name
                      ,'OriginalFilename', str(bootstrap_exe)
                      ,'version', self.app_version]
        win_app_package_exe_config.setVersionInfoInExe(
            str( self.package_folder / bootstrap_exe ),
            all_changes
            )

    vc_14_solution_file_template = '''Microsoft Visual Studio Solution File, Format Version 12.00
# Visual Studio 14
VisualStudioVersion = 14.0.23107.0
MinimumVisualStudioVersion = 10.0.40219.1
Project("{%(UUID1)s}") = "%(name)s", "%(exefile)s", "{%(UUID2)s}"
	ProjectSection(DebuggerProjectSystem) = preProject
		PortSupplier = 00000000-0000-0000-0000-000000000000
		Executable = %(Executable)s
		RemoteMachine = localhost
		StartingDirectory = %(StartingDirectory)s
		Environment = Default
		LaunchingEngine = 00000000-0000-0000-0000-000000000000
		UseLegacyDebugEngines = No
		LaunchSQLEngine = No
		AttachLaunchAction = No
	EndProjectSection
EndProject
Global
	GlobalSection(SolutionConfigurationPlatforms) = preSolution
		Release|x64 = Release|x64
	EndGlobalSection
	GlobalSection(ProjectConfigurationPlatforms) = postSolution
		{%(UUID2)s}.Release|x64.ActiveCfg = Release|x64
	EndGlobalSection
	GlobalSection(SolutionProperties) = preSolution
		HideSolutionNode = FALSE
	EndGlobalSection
EndGlobal
'''


class PackageFile:
    def __init__( self, app, source_file, package_file ):
        assert source_file.exists(), source_file
        assert not package_file.is_absolute(), package_file

        self.app = app
        self.source_file = source_file
        self.package_file = package_file

    def copy( self, msg_fmt, to_folder ):
        assert to_folder.relative_to( self.app.package_folder ), to_folder

        dest_file = to_folder / self.package_file
        dest_file.parent.mkdir( parents=True, exist_ok=True )

        self.__copy( msg_fmt, self.source_file, dest_file )

    def copyPdb( self, to_folder ):
        # if there is a .pdb file copy it
        source_pdb = self.source_file.with_suffix( '.pdb' )
        if not source_pdb.exists():
            return

        dest_pdb = (to_folder / self.package_file).with_suffix( '.pdb' )

        self.__copy( self.app.ct('Copy PDF <>em %s<>'), source_pdb, dest_pdb )

    def __copy( self, msg_fmt, src, dst ):
        self.app.verbose( msg_fmt % (src,) )
        self.app.verbose( self.app.ct('  to <>em %s<>') % (dst,) )

        with src.open( 'rb' ) as f_in, dst.open( 'wb' ) as f_out:
            while True:
                buf = f_in.read( 128 * 1024 )
                if len(buf) == 0:
                    break

                f_out.write( buf )

    def __repr__( self ):
        return '<PackageFile src=%s, pkg=%s>' % (self.source_file, self.package_file)

    # make sortable
    def __eq__( self, other ):
        if isinstance( other, PackageFile ):
            return self.source_file == other.source_file

        return self.source_file == other

    def __lt__( self, other ):
        if isinstance( other, PackageFile ):
            return self.source_file < other.source_file

        return self.source_file < other

    # make containable
    def __hash__( self ):
        return hash( self.source_file )



"""
setup.py for PythonWinAppPackager

"""

# Always prefer setuptools over distutils
import setuptools
import distutils.dist

import os.path

import win_app_packager
version = win_app_packager.VERSION

url = 'https://github.com/barry-scott/PythonWinAppPackager'

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

def getDevStatusFromVersion():
    if 'a' in version:
        return 'Development Status :: 3 - Alpha'

    elif 'b' in version:
        return 'Development Status :: 4 - Beta'

    else:
        return 'Development Status :: 5 - Production/Stable'

class DistributionAppPackager(distutils.dist.Distribution):
    # force wheel to be win32 or win64
    def has_c_libraries( self ):
        return True

setuptools.setup(
    name='win_app_packager',

    # force tag to include win32/win64 marker
    distclass=DistributionAppPackager, libraries = [],

    version=version,

    description='Python Win App Packager',
    long_description=long_description,
    long_description_content_type='text/x-rst',

    # The project's main homepage.
    url=url,

    # Author details
    author='Barry Scott',
    author_email='barry@barrys-emacs.org',

    # Choose your license
    license='Apache 2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        getDevStatusFromVersion(),

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # environment
        'Environment :: Win32 (MS Windows)',
        'Operating System :: Microsoft :: Windows :: Windows 7',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # What does your project relate to?
    keywords='development',

    packages=['win_app_packager'],

    package_data={'win_app_packager':
                        ['BootStrap/obj/bootstrap-cli.exe'
                        ,'BootStrap/obj/bootstrap-gui.exe']},

    install_requires=["namedstruct >=1.2.2", "colour_text"]
)
